from .SHS2Downloader import SHS2Downloader
from sentinelhub import DataCollection


class SHS2L2ADownloader(SHS2Downloader):

    def __init__(self, download_folder='/tmp'):
        super().__init__(download_folder)
        self.data_collection = DataCollection.SENTINEL2_L2A



import os
from dotenv import load_dotenv
import subprocess
import json
from pathlib import Path
import psutil


def load_status(project=None):
    try:
        with open(Path.home() / ".spai/status.json", "r") as f:
            status = json.load(f)
        if project is None:
            return status
        return status[project] if project in status else {}
    except:
        return {}


def save_status(status, project):
    os.makedirs(Path.home() / ".spai", exist_ok=True)
    prev_status = load_status()
    if status == {}:
        del prev_status[project]
    else:
        prev_status[project] = status
    with open(Path.home() / ".spai/status.json", "w") as f:
        json.dump(prev_status, f)


def kill_process(status, name, typer):
    process_id = status[name]
    try:
        process = psutil.Process(process_id)
        if process.is_running():
            typer.echo(f"Item '{name}' is already running. Stopping it...")
            # You can also obtain additional information about the process if needed
            # print(f"Process name: {process.name()}")
            # print(f"Process CPU usage: {process.cpu_percent()}%")
            # print(f"Process memory usage: {process.memory_info().rss} bytes")
            subprocess.call(["kill", str(process_id)])
        else:
            # print("Process has terminated.")
            del status[name]
    except psutil.NoSuchProcess:  # process has terminated
        # print("No such process")
        del status[name]


def run_local(dir, config, typer):
    typer.echo(f"Running locally...")
    project = config["project"]
    status = load_status(project)
    if "scripts" in config:
        typer.echo(f"Running scripts...")
        for script in config["scripts"]:
            # check if script is already running
            if script["name"] in status:
                kill_process(status, script["name"], typer)
            typer.echo(f"Running script '{script['name']}'...")
            # TODO: reqs, env
            os.chdir(f"{dir}/scripts/{script['name']}")
            # os.system(f"python main.py")
            # status[script["name"]] = "hola"
            # TODO: run in background and store job ids so we can kill them
            process = subprocess.Popen(
                ["python", "main.py"], shell=False, stdout=open(f"output.txt", "w")
            )
            process_id = process.pid
            print(f"Background process ID: {process_id}")
            status[script["name"]] = process_id
            save_status(status, project)
    if "notebooks" in config:
        typer.echo(f"Running notebooks...")
        for notebook in config["notebooks"]:
            typer.echo(f"Running notebook '{notebook['name']}'...")
            # TODO: reqs, env
            # papermill can execute notebooks and save output in S3
            # papermill can execute notebooks with params
            os.chdir(f"{dir}/notebooks/{notebook['name']}")
            os.system(f"papermill main.ipynb output.ipynb")
    if "apis" in config:
        typer.echo(f"Running apis...")
        for api in config["apis"]:
            # check if api is already running
            if api["name"] in status:
                kill_process(status, api["name"], typer)
            typer.echo(f"Running api '{api['name']}'...")
            # TODO: reqs, env
            load_dotenv(f"{dir}/apis/{api['name']}/.env")
            # change directory to api folder
            os.chdir(f"{dir}/apis/{api['name']}")
            # os.system(f"uvicorn main:app --port {api['port']}")
            process = subprocess.Popen(
                ["python", "main.py"],  # for this to work need init in main with port
                shell=False,
                stdout=open("output.txt", "w"),
            )
            process_id = process.pid
            print(f"Background process ID: {process_id}")
            status[api["name"]] = process_id
            # TODO: run in background so we can run multiple apis
            # os.system(f"uvicorn main:app --port {api['port']}")
            save_status(status, project)
    return


def stop_local(typer, project=None):
    if project is None:
        # TODO: stop all projects
        return
    status = load_status(project)
    if status == {}:
        typer.echo(f"No items running for project '{project}'.")
        return
    typer.echo(f"Stopping local...")
    delete = []
    for item, pid in status.items():
        try:
            process = psutil.Process(pid)
            if process.is_running():
                typer.echo(f"Stoppping item '{item}'...")
                subprocess.call(["kill", str(pid)])
            else:
                print(f"Item '{item}' has terminated.")
        except psutil.NoSuchProcess:
            print(f"Item '{item}' has terminated.")
        delete.append(item)
    for item in delete:
        del status[item]
    save_status(status, project)


def list_local(typer, project=None):
    status = load_status()
    if status == {}:
        typer.echo(f"No items running.")
        return
    for project, items in status.items():
        typer.echo(f"Project '{project}':")
        for item, pid in items.items():
            try:
                process = psutil.Process(pid)
                if process.is_running():
                    typer.echo(f"  Item '{item}' is running...")
                else:
                    print(f"  Item '{item}' has terminated.")
            except psutil.NoSuchProcess:
                print(f"  Item '{item}' has terminated.")
    return



# https://github.com/DHI-GRAS/terracotta/blob/main/terracotta/cache.py

from typing import Tuple, Callable, Any

import sys
import zlib

import numpy as np
from cachetools import LFUCache, Cache

CompressionTuple = Tuple[bytes, bytes, str, Tuple[int, int]]


class CompressedLFUCache(LFUCache):
    """Least-frequently-used cache with ZLIB compression"""

    def __init__(self, maxsize: int, compression_level: int):
        super().__init__(maxsize, self._get_size)
        self.compression_level = compression_level

    def __getitem__(
        self, key: Any, cache_getitem: Callable = Cache.__getitem__
    ) -> np.ma.MaskedArray:
        compressed_item = super().__getitem__(key, cache_getitem)
        return self._decompress_tuple(compressed_item)

    def __setitem__(
        self,
        key: Any,
        value: np.ma.MaskedArray,
        cache_setitem: Callable = Cache.__setitem__,
    ) -> None:
        val_compressed = self._compress_ma(value, self.compression_level)
        super().__setitem__(key, val_compressed, cache_setitem)

    @staticmethod
    def _compress_ma(
        arr: np.ma.MaskedArray, compression_level: int
    ) -> CompressionTuple:
        compressed_data = zlib.compress(arr.data, compression_level)
        mask_to_int = np.packbits(arr.mask.astype(np.uint8))
        compressed_mask = zlib.compress(mask_to_int, compression_level)
        out = (compressed_data, compressed_mask, arr.dtype.name, arr.shape)
        return out

    @staticmethod
    def _decompress_tuple(compressed_data: CompressionTuple) -> np.ma.MaskedArray:
        data_b, mask_b, dt, ds = compressed_data
        data = np.frombuffer(zlib.decompress(data_b), dtype=dt).reshape(ds)
        mask = np.frombuffer(zlib.decompress(mask_b), dtype=np.uint8)
        mask = np.unpackbits(mask)[: int(np.prod(ds))]
        mask = mask.reshape(ds)
        return np.ma.masked_array(data, mask=mask)

    @staticmethod
    def _get_size(x: Tuple) -> int:
        sizes = map(sys.getsizeof, x)
        return sum(sizes)



import mercantile
import rasterio
import cachetools
from rasterio import warp, windows, transform
from affine import Affine
from rasterio.vrt import WarpedVRT
import numpy as np
from matplotlib import cm
import threading
from .cache import CompressedLFUCache
from .errors import ImageOutOfBounds
from .image_utils import to_uint8, contrast_stretch
import shapely
import geopandas as gpd
import time
import os

_cache_lock = threading.RLock()

#: Size of raster file in-memory cache in bytes
RASTER_CACHE_SIZE: int = 1024 * 1024 * 1024  # 1 GB
#: Compression level of raster file in-memory cache, from 0-9
RASTER_CACHE_COMPRESS_LEVEL: int = 9

_raster_cache = CompressedLFUCache(
    RASTER_CACHE_SIZE, compression_level=RASTER_CACHE_COMPRESS_LEVEL
)

# esto solo funciona si bounds están en lat, lng !


def tile_exists(xyz, bounds):
    x, y, z = xyz
    t1 = mercantile.tile(bounds[0], bounds[3], z)
    t2 = mercantile.tile(bounds[2], bounds[1], z)
    return t1.x <= x <= t2.x and t1.y <= y <= t2.y


def get_tile_data(file_path, xyz, bands, tile_size=(256, 256)):
    # tic = time.perf_counter()
    x, y, z = xyz

    # Determine the destination tile and its mercator bounds
    dst_tile = mercantile.Tile(x=x, y=y, z=z)
    tile_bounds = mercantile.xy_bounds(dst_tile)

    # print(file_path)

    src = rasterio.open(file_path)

    # toc = time.perf_counter()
    # print(f"1 in {toc - tic:0.8f} seconds")

    # convert src bounds to lat,lng
    # tic = time.perf_counter()
    bb = shapely.geometry.box(*src.bounds)
    # toc = time.perf_counter()
    # print(f"2 in {toc - tic:0.8f} seconds")

    # tic = time.perf_counter()
    gdf = gpd.GeoDataFrame(index=[0], crs=src.meta["crs"], geometry=[bb]).to_crs(
        "epsg:4326"
    )
    # toc = time.perf_counter()
    # print(f"3 in {toc - tic:0.8f} seconds")

    # tic = time.perf_counter()
    # check if tile is inside image bounds
    if not tile_exists(xyz, gdf.geometry.iloc[0].bounds):
        raise ImageOutOfBounds(x, y, z)
    # toc = time.perf_counter()
    # print(f"4 in {toc - tic:0.8f} seconds")

    # tic = time.perf_counter()
    # return cached tile if exists
    kwargs = dict(
        path=file_path.split("?")[0],
        tile_bounds=tuple(tile_bounds) if tile_bounds else None,
        tile_size=tuple(tile_size),
        bands=bands,
        # preserve_values=preserve_values,
        # reprojection_method=settings.REPROJECTION_METHOD,
        # resampling_method=settings.RESAMPLING_METHOD
    )
    cache_key = cachetools.keys.hashkey(**kwargs)
    try:
        with _cache_lock:
            return _raster_cache[cache_key]
    except KeyError:
        pass
    # toc = time.perf_counter()
    # print(f"5 in {toc - tic:0.8f} seconds")

    # vrt = WarpedVRT(src, crs='EPSG:3857', resampling=Resampling.bilinear)

    # tic = time.perf_counter()
    # compute suggested resolution in target CRS
    dst_transform, _, _ = warp.calculate_default_transform(
        src.crs, "EPSG:3857", src.width, src.height, *src.bounds
    )
    dst_res = (abs(dst_transform.a), abs(dst_transform.e))
    # toc = time.perf_counter()
    # print(f"6 in {toc - tic:0.8f} seconds")

    # make sure VRT resolves the entire tile
    # tile_transform = transform.from_bounds(*tile_bounds, *tile_size)
    # tile_res = (abs(tile_transform.a), abs(tile_transform.e))

    # tic = time.perf_counter()
    # pad tile bounds to prevent interpolation artefacts
    num_pad_pixels = 2

    # compute tile VRT shape and transform
    dst_width = max(1, round((tile_bounds[2] - tile_bounds[0]) / dst_res[0]))
    dst_height = max(1, round((tile_bounds[3] - tile_bounds[1]) / dst_res[1]))
    vrt_transform = transform.from_bounds(
        *tile_bounds, width=dst_width, height=dst_height
    ) * Affine.translation(-num_pad_pixels, -num_pad_pixels)
    vrt_height, vrt_width = (
        dst_height + 2 * num_pad_pixels,
        dst_width + 2 * num_pad_pixels,
    )
    # toc = time.perf_counter()
    # print(f"7 in {toc - tic:0.8f} seconds")

    # tic = time.perf_counter()
    # remove padding in output
    window = windows.Window(
        col_off=num_pad_pixels,
        row_off=num_pad_pixels,
        width=dst_width,
        height=dst_height,
    )
    # toc = time.perf_counter()
    # print(f"8 in {toc - tic:0.8f} seconds")

    # tic = time.perf_counter()
    # construct VRT
    vrt = WarpedVRT(
        src,
        crs="EPSG:3857",
        transform=vrt_transform,
        width=vrt_width,
        height=vrt_height,
    )
    # toc = time.perf_counter()
    # print(f"9 in {toc - tic:0.8f} seconds")

    # EL PROBLEMA PARECE ESTAR AQUÍ
    # aunque el timer da poco tiempo, la request se queda colgada un buen rato
    # tic = time.perf_counter()
    # for i in range(1000):
    tile_data = vrt.read(bands, window=window, out_shape=tile_size)
    # toc = time.perf_counter()
    # print(f"10 in {toc - tic:0.8f} seconds")

    # tic = time.perf_counter()
    # assemble alpha mask
    # mask_idx = vrt.count
    # mask = vrt.read(mask_idx, window=window, out_shape=self.tile_size) == 0
    mask = vrt.read(bands, window=window, out_shape=tile_size) == 0
    # toc = time.perf_counter()
    # print(f"11 in {toc - tic:0.8f} seconds")

    if src.nodata is not None:
        mask |= tile_data == src.nodata

    # tic = time.perf_counter()
    result = np.ma.masked_array(tile_data, mask=mask)
    # toc = time.perf_counter()
    # print(f"12 in {toc - tic:0.8f} seconds")

    # add to cache
    try:
        with _cache_lock:
            _raster_cache[cache_key] = result
    except ValueError:  # value too large
        pass

    return result


def get_image_data(tile_data, stretch, palette):
    # estoy hay que hacerlo mejor
    if tile_data.ndim == 2:
        # nodes_colors = palette.split(';') # 1,2,3;color1,color2,color3
        # if nodes_colors.length == 1:
        if palette:
            # pasamos de min-max al rango determinado por stretch (0-1)
            tile_data = contrast_stretch(
                tile_data, (stretch[0], stretch[1]), (0.0, 1.0), clip=True
            )
            # aplicar paleta (0-1)
            img_data = getattr(cm, palette)(tile_data)
            # convertir a uint8
            img_data = (img_data * 255).astype(np.uint8)
            return img_data
        # elif nodes_colors.length == 2:
        # nodes, colors = nodes_colors[0], nodes_colors[1]
        # assert len(nodes) == len(colors), 'nodes and colors must have the same length'
        # nodes = [float(n) for n in nodes.split(',')]
        # colors = colors.split(',)
        # apply palette
        else:
            tile_data = np.ma.stack([tile_data] * 3).transpose(1, 2, 0)
            return to_uint8(tile_data, stretch[0], stretch[1])
    elif tile_data.ndim == 3:
        tile_data = tile_data.transpose(1, 2, 0)
        return to_uint8(tile_data, stretch[0], stretch[1])
    else:
        raise ValueError("wrong dims")



import typer
from pathlib import Path
import os
import shutil
import sys

# Add the cli directory to the Python path
spai_cli_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(spai_cli_dir))

from cli.commands import hello
from cli import (
    load_and_validate_config,
    deploy_and_run_cloud,
    run_local,
    stop_local,
    list_local,
)

app = typer.Typer()
app.add_typer(hello.app, name="hello")


@app.command()
def init():
    # ask for project name
    project = typer.prompt("Project name")
    # copy template
    template = Path(__file__).parent / "cli" / "project-template"
    shutil.copytree(template, Path(project))
    # change name to project in spai.config.yaml
    config = Path(project) / "spai.config.yml"
    os.system(f"sed -i 's/project-template/{project}/g' {config}")
    return typer.echo(f"Project {project} created")


@app.command()
def run(
    dir: Path = typer.Option(Path("."), "-d"),
    cloud: bool = typer.Option(False, "-c", "--cloud"),
):
    # make sure dir is absolute
    dir = Path(dir).resolve()
    # load and validate config
    config = load_and_validate_config(dir, typer, cloud)
    # print(config)
    # run project
    if cloud:
        return deploy_and_run_cloud(dir, config, typer)
    return run_local(dir, config, typer)


@app.command()
def list(
    project: str = typer.Option(None, "-p", "--project"),
):
    return list_local(typer, project)


@app.command()
def stop(
    project: str = typer.Option(None, "-p", "--project"),
):
    # TODO: get confirmation
    return stop_local(typer, project)


if __name__ == "__main__":
    app()



import os
import yaml


def validate_folder(dir, folder, typer):
    if not os.path.exists(dir / folder):
        raise typer.BadParameter(f"No {folder} directory found in '{dir}'.")


def validate_item(dir, folder, item, name, typer, file="main.py"):
    # check name
    if not "name" in item:
        raise typer.BadParameter(
            f"{name} '{item['name']}' is missing 'name' attribute."
        )
    # check folder has folder with item name
    if not os.path.exists(dir / folder / item["name"]):
        raise typer.BadParameter(
            f"{name} '{item['name']}' cannot be found in {dir}/{folder}."
        )
    # check folder has file
    if not file in os.listdir(dir / folder / item["name"]):
        raise typer.BadParameter(f"{name} '{item['name']}' is missing file 'main.py'.")
    # TODO: check optionals: reqs, env...


def load_and_validate_config(dir, typer, cloud):
    # check dir exists
    if not dir.exists():
        raise typer.BadParameter(f"Directory '{dir}' does not exist.")
    # check dir is a spai project
    if not "spai.config.yml" in os.listdir(dir):
        raise typer.BadParameter(
            f"Directory '{dir}' is not a spai project. No spai.config.yml file found."
        )
    # load config
    config = {}
    with open(dir / "spai.config.yml", "r") as f:
        config = yaml.safe_load(f)
    # check config has project name if cloud deployment
    if not "project" in config:
        raise typer.BadParameter(f"spai.config.yml file is missing 'project' section.")
    # check scripts
    if "scripts" in config:
        # check project has scripts folder
        validate_folder(dir, "scripts", typer)
        for script in config["scripts"]:
            validate_item(dir, "scripts", script, "script", typer)
    # check notebooks
    if "notebooks" in config:
        # check project has notebooks folder
        validate_folder(dir, "notebooks", typer)
        for notebook in config["notebooks"]:
            validate_item(dir, "notebooks", notebook, "notebook", typer, "main.ipynb")
    # check apis
    if "apis" in config:
        # check project has apis folder
        validate_folder(dir, "apis", typer)
        for api in config["apis"]:
            validate_item(dir, "apis", api, "api", typer)
    # check uis
    if "uis" in config:
        # check project has uis folder
        validate_folder(dir, "uis", typer)
        for ui in config["uis"]:
            validate_item(dir, "uis", ui, "ui", typer)
    return config



import os
from sentinelhub import SHConfig, SentinelHubCatalog
from sentinelhub import CRS, BBox
from sentinelhub.data_collections import DataCollection


class SHExplorer:
    def __init__(self, time_interval, sensor, **kargs):
        self.query, self.fields = None, {
            "include": ["id", "properties.datetime", "assets.thumbnail.href"],
            "exclude": [],
        }
        self.config = SHConfig()
        if sensor == "S1":
            self.data_collection = DataCollection.SENTINEL1
        elif sensor == "S2L1C" or sensor == "S2L2A":
            # no thumbnail for S2L2A, use the one for S2L1C
            self.data_collection = DataCollection.SENTINEL2_L1C
            if "cloud_cover" in kargs:
                cloud_cover = kargs["cloud_cover"]
                self.query = f"eo:cloud_cover < {cloud_cover}"
                self.fields["include"].append("properties.eo:cloud_cover")
        elif sensor == "S5P":
            self.data_collection = DataCollection.SENTINEL5P
            # self.data_collection = DataCollection.SENTINEL3_OLCI
            self.config.sh_base_url = DataCollection.SENTINEL5P.service_url
            # self.config.sh_base_url = DataCollection.SENTINEL3_OLCI.service_url
        else:
            raise Exception("Invalid sensor")
        self.time_interval = time_interval
        self.config.sh_client_id = os.environ["SH_CLIENT_ID"]
        self.config.sh_client_secret = os.environ["SH_CLIENT_SECRET"]
        self.config.save()

    def search(self, gdf):
        # generate the bbox containing the geometry
        mybbox = BBox(bbox=gdf.total_bounds.tolist(), crs=CRS.WGS84)
        # query (different for each server: pleiades, spot, sentinelhub...)
        catalog = SentinelHubCatalog(config=self.config)
        search_iterator = catalog.search(
            self.data_collection,
            bbox=mybbox,
            time=self.time_interval,
            filter=self.query,
            fields=self.fields,
        )
        _results = list(search_iterator)
        results = []
        for result in _results:
            data = {
                "id": result["id"],
                "thumbnail": result["assets"]["thumbnail"]["href"]
                if "assets" in result
                else None,
                "date": result["properties"]["datetime"],
            }
            if "eo:cloud_cover" in result["properties"]:
                data["cloud_cover"] = result["properties"]["eo:cloud_cover"]
            results.append(data)
        return results



import typer

app = typer.Typer()


@app.command()
def hello(name: str = 'World!'):
    """Say hello to someone"""
    typer.echo(f"hello, {name}")

@app.command()
def goodbye(name: str = 'World!'):
    """Say goodbye to someone"""
    typer.echo(f"goodbye, {name}")


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spai',
 'spai.cli',
 'spai.cli.commands',
 'spai.cli.project-template.apis.analytics',
 'spai.cli.project-template.apis.xyz',
 'spai.cli.project-template.scripts.downloader',
 'spai.cli.project-template.scripts.ndvi',
 'spai.cli.project-template.uis.map',
 'spai.data',
 'spai.data.satellite',
 'spai.data.satellite.sentinelhub',
 'spai.image',
 'spai.image.xyz',
 'spai.storage']

package_data = \
{'': ['*'],
 'spai.cli': ['project-template/*', 'project-template/notebooks/analytics/*']}

entry_points = \
{'console_scripts': ['spai = spai.main:app']}

setup_kwargs = {
    'name': 'spai',
    'version': '2023.5.25.post6',
    'description': '',
    'long_description': '',
    'author': 'Juan Sensio',
    'author_email': 'it@earthpulse.es',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import rasterio as rio
import numpy as np
from PIL import Image


# TODO: should know the source in order to generate the rgb thumbnail
def thumbnail(image, size=None, factor=1):
    sensor = image.split("/")[-1].split("_")[0]
    if sensor == "S2L2A" or sensor == "S2L1C":
        ds = rio.open(image)
        rgb = ds.read((4, 3, 2)) / 4000
        rgb = np.clip(rgb, 0, 1)
        rgb = np.moveaxis(rgb, 0, -1)
        rgb = (rgb * 255).astype(np.uint8)
        image = Image.fromarray(rgb)
    elif sensor == "S1":
        ds = rio.open(image)
        vv = ds.read(1)
        vh = ds.read(2)
        # vv = np.clip(2.0 * vv, 0, 1)
        rgb = np.stack([(5.5 * vh > 0.5).astype(float), vv, 8 * vh], axis=-1)
        # db = 10 * np.log10(vv)
        # print(db.shape, db.min(), db.max())
        # db = np.clip(db, -30, 0) * (-8.4)
        # print(db.shape, db.min(), db.max())
        rgb = np.clip(rgb, 0, 1)
        rgb = (rgb * 255.0).astype(np.uint8)
        image = Image.fromarray(rgb)
    else:
        raise Exception(f"sensor {sensor} not supported")
    if size:
        image.thumbnail(size)
    elif factor != 1:
        assert factor > 0, "factor must be greater than 0"
        return image.resize((int(image.width // factor), int(image.height // factor)))
    return image
